# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from webob.exc import HTTPForbidden
from nose.tools import istest
from reahl.tofu import test, Fixture, scenario, expected, vassert
from reahl.stubble import stubclass, EmptyStub

from reahl.webdev.tools import WidgetTester, Browser, XPath
from reahl.web_dev.fixtures import WebFixture
from reahl.web.fw import Widget, UserInterface
from reahl.web.ui import Form, TextInput, ButtonInput, Button, Panel, P, TwoColumnPage
from reahl.component.modelinterface import Field, AccessRights, Event, exposed, Allowed, Action


    
@istest
class SecurityTests(object):
    @test(WebFixture)
    def security_sensitive(self, fixture):
        """A Widget is security sensitive if a read_check is specified for it; one of its
           children is security sensitive, or it was explixitly marked as security sensitive."""
        
        widget = Widget(fixture.view)

        # Case: marked explicitly        
        vassert( not widget.is_security_sensitive )
        widget.set_as_security_sensitive()
        vassert( widget.is_security_sensitive )

        # Case: derived from children
        parent_widget = Widget(fixture.view)
        parent_widget.add_child(widget)        
        vassert( widget.is_security_sensitive )

        # Case: derived from read rights
        def allow(self): return True
        widget_with_rights = Widget(fixture.view, read_check=allow)
        vassert( widget.is_security_sensitive )

    @test(WebFixture)
    def serving_security_sensitive_widgets(self, fixture):
        """If the page is security sensitive, it will only be served on config.web.encrypted_http_scheme,
           else it will only be served on config.web.default_http_scheme."""
        class TestPanel(Panel):
            def __init__(self, view):
                super(TestPanel, self).__init__(view)
                widget = self.add_child(Widget(view))

                if fixture.security_sensitive:
                    widget.set_as_security_sensitive()

        wsgi_app = fixture.new_wsgi_app(child_factory=TestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        vassert( fixture.config.web.encrypted_http_scheme == u'https' )
        vassert( fixture.config.web.default_http_scheme == u'http' )

        fixture.security_sensitive = False
        fixture.driver_browser.open(u'https://localhost:8363/')
        vassert( fixture.driver_browser.current_url.scheme == fixture.config.web.default_http_scheme )
        
        fixture.security_sensitive = True
        fixture.driver_browser.open(u'http://localhost:8000/')
        vassert( fixture.driver_browser.current_url.scheme == fixture.config.web.encrypted_http_scheme )
        
    @test(Fixture)
    def fields_have_access_rights(self, fixture):
        """Fields have access rights for reading and for writing. By default both reading and writing are allowed.
           This default can be changed by passing an Action (which returns a boolean) for each kind of
           access which will be called to determine whether that kind of access is allowed.
           """

        def not_allowed(): return False

        # Case: the default
        field = Field()
        vassert( field.can_read() )
        vassert( field.can_write() )

        # Case: tailored rights
        field = Field(readable=Action(not_allowed), writable=Action(not_allowed))
        field.bind(u'field_name', fixture)
        vassert( not field.can_read() )
        vassert( not field.can_write() )
        
    @test(WebFixture)
    def tailored_access_make_inputs_security_sensitive(self, fixture):
        """An Input is sensitive if explicitly set as sensitive, or if its Fields has non-defaulted
           mechanisms for determiing access rights."""

        form = Form(fixture.view, u'some_form')
        field = Field(default=3, readable=Allowed(True))
        field.bind(u'field_name', EmptyStub())
        input_widget = TextInput(form, field)

        vassert( input_widget.is_security_sensitive )

    class InputRenderingScenarios(WebFixture):
        def new_form(self):
            return Form(self.view, u'some_form')
        def new_field(self):
            field = Field(default=3, readable=self.readable, writable=self.writable)
            field.bind(u'field_name', EmptyStub())
            return field
        def new_event(self):
            event = Event(readable=self.readable, writable=self.writable)
            event.bind(u'event_name', EmptyStub())
            self.form.define_event_handler(event)
            return event

        @scenario
        def normal_rendering(self):
            self.readable = Allowed(True)
            self.writable = Allowed(True)
            self.input_widget = TextInput(self.form, self.field)

            self.expected_html = u'<input name="field_name" form="some_form" type="text" value="3" class="reahl-textinput">'

        @scenario
        def disabled_rendering(self):
            self.readable = Allowed(True)
            self.writable = Allowed(False)
            self.input_widget = TextInput(self.form, self.field)

            self.expected_html = u'<input name="field_name" disabled="disabled" form="some_form" type="text" value="3" class="reahl-textinput">'

        @scenario
        def valueless_rendering(self):
            self.readable = Allowed(False)
            self.writable = Allowed(True)
            self.input_widget = TextInput(self.form, self.field)

            self.expected_html = u'<input name="field_name" form="some_form" type="text" value="" class="reahl-textinput">'
        
        @scenario
        def empty_rendering(self):
            self.readable = Allowed(False)
            self.writable = Allowed(False)
            self.input_widget = TextInput(self.form, self.field)

            self.expected_html = u''
            
        @scenario
        def normal_button_rendering(self):
            self.readable = Allowed(True)
            self.writable = Allowed(True)
            self.input_widget = ButtonInput(self.form, self.event)

            self.expected_html = u'<input name="event.event_name?" form="some_form" type="submit" value="event_name">'

        @scenario
        def greyed_button_rendering(self):
            self.readable = Allowed(True)
            self.writable = Allowed(False)
            self.input_widget = ButtonInput(self.form, self.event)

            self.expected_html = u'<input name="event.event_name?" disabled="disabled" form="some_form" type="submit" value="event_name">'

        @scenario
        def buttons_must_be_readable_to_be_present(self):
            self.readable = Allowed(False)
            self.writable = Allowed(True)
            self.input_widget = ButtonInput(self.form, self.event)

            self.expected_html = u''

        @scenario
        def nothing_allowed_on_button(self):
            self.readable = Allowed(False)
            self.writable = Allowed(False)
            self.input_widget = ButtonInput(self.form, self.event)

            self.expected_html = u''

        def allowed(self):
            return True
        def disallowed(self):
            return False

        @scenario
        def not_readable_widget(self):
            self.input_widget = Widget(self.view, read_check=self.disallowed)
            self.input_widget.add_child(P(self.view, text=u'some text in a p'))

            self.expected_html = u''

        @scenario
        def readable_but_not_writable_widget(self):
            self.input_widget = Widget(self.view, read_check=self.allowed, write_check=self.disallowed)
            self.input_widget.add_child(P(self.view, text=u'some text in a p'))

            self.expected_html = u'<p>some text in a p</p>'

        @scenario
        def readable_and_writable_widget(self):
            self.input_widget = Widget(self.view, read_check=self.allowed, write_check=self.allowed)
            self.input_widget.add_child(P(self.view, text=u'some text in a p'))

            self.expected_html = u'<p>some text in a p</p>'

        @scenario
        def not_readable_but_writable_widget(self):
            # Think of password field, where the input's value is not readable by a human
            self.input_widget = Widget(self.view, read_check=self.disallowed, write_check=self.allowed)
            self.input_widget.add_child(P(self.view, text=u'some text in a p'))

            self.expected_html = u''


    @test(InputRenderingScenarios)
    def rendering_inputs(self, fixture):
        """How Inputs render, depending on the rights."""

        tester = WidgetTester(fixture.input_widget)
        
        actual = tester.render_html()
        vassert( actual == fixture.expected_html )


    @test(WebFixture)
    def non_writable_input_is_dealt_with_like_invalid_input(self, fixture):
        """If a form submits a value for an Input that is linked to Field with access rights that prohibit writing,
           the input is silently ignored."""

        class ModelObject(object):
            field_name = u'Original value'
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me')

            @exposed
            def fields(self, fields):
                fields.field_name = Field(default=u'abc', writable=Allowed(False), disallowed_message=u'you are not allowed to write this')
        model_object = ModelObject()

        class TestPanel(Panel):
            def __init__(self, view):
                super(TestPanel, self).__init__(view)
                form = self.add_child(Form(view, u'some_form'))
                form.define_event_handler(model_object.events.an_event)
                form.add_child(Button(form, model_object.events.an_event))
                form.add_child(TextInput(form, model_object.fields.field_name))
                fixture.form = form

        wsgi_app = fixture.new_wsgi_app(child_factory=TestPanel.factory())
        browser = Browser(wsgi_app)
        browser.open(u'/')

        browser.post(fixture.form.event_channel.get_url().path, {u'event.an_event?':u'', u'field_name': 'illigitimate value'})
        browser.follow_response()
        vassert( model_object.field_name == u'Original value' )


    @test(WebFixture)
    def non_writable_events_are_dealt_with_like_invalid_input(self, fixture):
        """If a form submits an Event with access rights that prohibit writing, a ValidationException is raised."""

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', writable=Allowed(False), 
                                        disallowed_message=u'you cannot do this')

        model_object = ModelObject()
        class TestPanel(Panel):
            def __init__(self, view):
                super(TestPanel, self).__init__(view)
                form = self.add_child(Form(view, u'some_form'))
                form.define_event_handler(model_object.events.an_event)
                form.add_child(Button(form, model_object.events.an_event))
                fixture.form = form

        wsgi_app = fixture.new_wsgi_app(child_factory=TestPanel.factory())
        browser = Browser(wsgi_app)
        browser.open(u'/')

        browser.post(fixture.form.event_channel.get_url().path, {u'event.an_event?':u''})
        browser.follow_response()
        error_label = browser.get_html_for('//label')
        vassert( error_label == u'<label for="event.an_event?" class="error">you cannot do this</label>' )

    @test(WebFixture)
    def getting_view(self, fixture):
        """ONLY If a View is readable, it can be GET"""
        def disallowed(): return False
        
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_view(u'/view', u'Title', read_check=disallowed)
        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        browser.open(u'/view', status=403)

    @test(WebFixture)
    def posting_to_view(self, fixture):
        """ONLY If a View is writable, may it be POSTed to"""
        def disallowed(): return False
        
        class MyForm(Form):
            def __init__(self, view):
                super(MyForm, self).__init__(view, u'myform')
                self.define_event_handler(self.events.an_event)
                self.add_child(Button(self, self.events.an_event))
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'Click me')

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/a_view', u'Title', write_check=disallowed)
                home.set_slot(u'main', MyForm.factory())
        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open(u'/a_view')
        browser.click(XPath.button_labelled(u'Click me'), status=403)

    






# Copyright 2010-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.stubble import stubclass
from nose.tools import istest
from reahl.tofu import Fixture, test, scenario, set_up
from reahl.tofu import vassert

from reahl.webdev.tools import XPath
from reahl.web.fw import Bookmark
from reahl.web.ui import P, Div, HTML5Page, Panel
from reahl.web.pager import Page, PageIndex, PageMenu, PagedPanel
from reahl.web_dev.fixtures import WebBasicsMixin
from reahl.component.context import ExecutionContext

from reahl.web.pager import SequentialPageIndex, AnnualPageIndex, AnnualItemOrganiserProtocol

class PageMenuFixture(Fixture, WebBasicsMixin):
    def new_number_of_pages(self):
        return 10
        
    def new_max_page_links(self):
        return 5
        
    def new_PageIndexStub(self):
        fixture = self
        @stubclass(PageIndex)
        class PageIndexStub(PageIndex):
            def __init__(self, max_page_links, number_of_pages):
                super(PageIndexStub, self).__init__(max_page_links=max_page_links)
                self.number_of_pages = number_of_pages

            @property
            def total_number_of_pages(self):
                return self.number_of_pages
            
            def get_contents_for_page(self, page_number):
                return u'contents of page %s' % page_number

            def get_description_for_page(self, page_number):
                return u'p%s' % page_number

        return PageIndexStub
        
    def new_PageContainer(self):
        class PageContainer(PagedPanel):
            def __init__(self, parent, page_index):
                super(PageContainer, self).__init__(parent, page_index, u'container')
                self.add_child(P(self.view, text=self.current_contents))

        return PageContainer

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Panel):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)
                page_index = fixture.PageIndexStub(fixture.max_page_links, fixture.number_of_pages)
                page_container = self.add_child(fixture.PageContainer(self.view, page_index))
                self.add_child(PageMenu(self.view, u'page_menu_widget', page_index, page_container))
        return MainWidget
        
    def container_contents_is(self, expected):
        return self.driver_browser.execute_script('return window.jQuery("div p").html() == "%s"' % expected)

    def page_range_links_match(self, link_labels):
        return self.driver_browser.execute_script('return window.jQuery(".reahl-pagemenu a").slice(2,-2).map(function(){return window.jQuery(this).html();}).toArray() == "%s"' % link_labels)

    def new_wsgi_app(self):
        return super(PageMenuFixture, self).new_wsgi_app(enable_js=True, 
                                                       child_factory=self.MainWidget.factory())


@istest
class PageMenuTests(object):
    # Please see fixture for how to declare a PageIndex, PageContainer and PageMenu
    @test(PageMenuFixture)
    def selecting_a_page(self, fixture):
        """Clicking the link of a page results in the contents of the PageContainer being refreshed."""
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open(u'/')

        fixture.driver_browser.wait_for(fixture.container_contents_is, u'contents of page 1')
        fixture.driver_browser.click(XPath.link_with_text(u'p2'))
        fixture.driver_browser.wait_for(fixture.container_contents_is, u'contents of page 2')

    @test(PageMenuFixture)
    def navigating_the_page_numbers(self, fixture):
        """One can navigate the range of page links displayed by the PageMenu using the special links."""
        fixture.number_of_pages = 30
        fixture.max_page_links = 5
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open(u'/')

        # Case: next link
        fixture.driver_browser.click(XPath.link_with_text(u'>'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, u'p6,p7,p8,p9,p10') )
        fixture.driver_browser.click(XPath.link_with_text(u'>'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, u'p11,p12,p13,p14,p15') )

        # Case: prev link
        fixture.driver_browser.click(XPath.link_with_text(u'<'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, u'p6,p7,p8,p9,p10') )

        # Case: last link
        fixture.driver_browser.click(XPath.link_with_text(u'>|'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, u'p26,p27,p28,p29,p30') )

        # Case: first link
        fixture.driver_browser.click(XPath.link_with_text(u'|<'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, u'p1,p2,p3,p4,p5') )


    @test(PageMenuFixture)
    def active_state_of_next_prev_links(self, fixture):
        """Next and Last links are only active when not on the last range of pages,
           and Prev and First are similarly deactive when on the first range of pages."""
        fixture.number_of_pages = 15
        fixture.max_page_links = 5
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open(u'/')

        # Case: when you are on the left of the page range        
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text(u'|<')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text(u'<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'>')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'>|')) )

        # Case: when you are in the middle of the page range        
        fixture.driver_browser.click(XPath.link_with_text(u'>'))
        fixture.driver_browser.wait_for_element_present(XPath.link_with_text(u'p6'))
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'|<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'>')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'>|')) )
        
        # Case: when you are at the end of the page range        
        fixture.driver_browser.click(XPath.link_with_text(u'>'))
        fixture.driver_browser.wait_for_element_present(XPath.link_with_text(u'p11'))
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'|<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text(u'<')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text(u'>')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text(u'>|')) )

    class LinkScenarios(PageMenuFixture):
        @scenario
        def only_one_page(self):
            self.number_of_pages = 1
            self.max_page_links = 5
            self.goto_last_range = False
            self.visible_page_descriptions = u'p1'
            self.visible_last_page_descriptions = self.visible_page_descriptions

        @scenario
        def fewer_pages_than_max_links(self):
            self.number_of_pages = 2
            self.max_page_links = 5
            self.goto_last_range = False
            self.visible_page_descriptions = u'p1,p2'
            self.visible_last_page_descriptions = self.visible_page_descriptions

        @scenario
        def more_pages_than_max_links(self):
            self.number_of_pages = 10
            self.max_page_links = 5
            self.goto_last_range = True
            self.visible_page_descriptions = u'p1,p2,p3,p4,p5'
            self.visible_last_page_descriptions = u'p6,p7,p8,p9,p10'

        @scenario
        def more_than_one_range_last_range_not_complete(self):
            self.number_of_pages = 7
            self.max_page_links = 5
            self.goto_last_range = True
            self.visible_page_descriptions = u'p1,p2,p3,p4,p5'
            self.visible_last_page_descriptions = u'p3,p4,p5,p6,p7'

        @scenario
        def changing_number_of_links(self):
            self.number_of_pages = 10
            self.max_page_links = 3
            self.goto_last_range = False
            self.visible_page_descriptions = u'p1,p2,p3'
            self.visible_last_page_descriptions = self.visible_page_descriptions

    @test(LinkScenarios)
    def which_links_display(self, fixture):
        """The menu displayes the correct range of page links, depending on the starting page in the range, the
           total number of pages and the max number of links in a range"""
        fixture.reahl_server.set_app(fixture.wsgi_app)

        fixture.driver_browser.open(u'/')
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, fixture.visible_page_descriptions) )

        if fixture.goto_last_range:
            fixture.driver_browser.click(XPath.link_with_text(u'>|'))

        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, fixture.visible_last_page_descriptions) )

    @test(Fixture)
    def annual_page_index(self, fixture):
        """The AnnualPageIndex breaks a query of items up into pages containing items with the same year."""
        
        @stubclass(AnnualItemOrganiserProtocol)
        class ItemOrganiser(AnnualItemOrganiserProtocol):
            def get_years(self):                return [2000, 2001]
            def get_items_for_year(self, year): return [year]*3

        page_index = AnnualPageIndex(ItemOrganiser())

        vassert( page_index.get_contents_for_page(2) == [2001, 2001, 2001] )
        vassert( page_index.get_description_for_page(1) == u'2000' )
        vassert( page_index.years == [2000, 2001] )
        vassert( page_index.total_number_of_pages == 2 )


    class SequentialScenarios(Fixture):
        def new_items(self):
            return range(0, self.number_of_items)

        def new_page_index(self):
            return SequentialPageIndex(self.items, items_per_page=self.items_per_page, max_page_links=12)
            
        @scenario
        def less_than_a_page_items(self):
            self.number_of_items = 2
            self.items_per_page = 10
            self.expected_pages = 1
            self.last_page_contents = self.items[0:2]

        @scenario
        def exactly_one_page_items(self):
            self.number_of_items = 10
            self.items_per_page = 10
            self.expected_pages = 1
            self.last_page_contents = self.items[0:10]

        @scenario
        def just_more_than_one_page_items(self):
            self.number_of_items = 15
            self.items_per_page = 10
            self.expected_pages = 2
            self.last_page_contents = self.items[10:15]
        
    @test(SequentialScenarios)
    def sequential_page_index(self, fixture):
        """The SequentialPageIndex breaks a query of items up into pages of sequential items"""
        
        vassert( fixture.page_index.total_number_of_pages == fixture.expected_pages )
        page = fixture.page_index.get_page_number(fixture.expected_pages)
        vassert( page.description == unicode(fixture.expected_pages) )
        vassert( page.contents == fixture.last_page_contents )        
        vassert( fixture.page_index.max_page_links == 12 )





# Copyright 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os.path
import threading 
import functools
import socket

from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.stubble import EmptyStub, stubclass
from reahl.tofu import vassert, expected, NoException, temp_file_with

from reahl.component.context import ExecutionContext
from reahl.component.exceptions import DomainException
from reahl.component.modelinterface import FileField, exposed, Event, Action, ValidationConstraint
from reahl.web.ui import SimpleFileInput, FileUploadInput, FileUploadPanel, Button, Form
from reahl.web.fw import UploadedFile
from reahl.web_dev.fixtures import WebFixture
from reahl.webdev.tools import Browser, XPath
from reahl.webelixirimpl import PersistedFile


class FileUploadInputFixture(WebFixture):
    def file_was_uploaded(self, filename):
        return PersistedFile.query.filter_by(filename=os.path.basename(filename)).count() == 1

    file_to_upload1_name = u'file1.html'
    file_to_upload2_name = u'file2.gif'
    file_to_upload1_content = u'some content'
    file_to_upload2_content = u'some different content'

    def new_file_to_upload1(self):
        return temp_file_with(self.file_to_upload1_content, name=self.file_to_upload1_name)

    def new_file_to_upload2(self):
        return temp_file_with(self.file_to_upload2_content, name=self.file_to_upload2_name)

    def new_domain_object(self):
        class DomainObject(object):
            def __init__(self):
                self.throws_exception = False
                self.files = []
                self.submitted_file_info = {}
                self.submitted = False
               
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label=u'Attached files')

            @exposed
            def events(self, events):
                events.submit = Event(label=u'Submit', action=Action(self.submit))
                
            def submit(self):
                if self.throws_exception:
                    raise DomainException()
                for f in self.files:
                    with f.open() as opened_file:
                        contents = opened_file.read()
                    self.submitted_file_info[f.filename] = (contents, f.content_type)
                self.submitted = True
                
        return DomainObject()

    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, u'test')
                self.add_child(FileUploadInput(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm

    def new_wsgi_app(self, enable_js=False):
        return super(FileUploadInputFixture, self).new_wsgi_app(child_factory=self.FileUploadForm.factory(), enable_js=enable_js)

    def uploaded_file_is_listed(self, filename):
        return self.driver_browser.is_element_present('//ul/li/span[text()="%s"]' % os.path.basename(filename))


class ConstrainedFileUploadInputFixture(FileUploadInputFixture):
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = fixture.file_field

            @exposed
            def events(self, events):
                events.submit = Event(label=u'Submit')

        return DomainObject()

class PerFileConstrainedFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    @scenario
    def size_constrained(self):
        max_size = 13
        self.file_field = FileField(allow_multiple=True, max_size_bytes=max_size, label=u'Attached files')
        self.validation_error_message = u'files should be smaller than 13.0bytes'
        self.valid_file = temp_file_with(u'c'*max_size, name=u'valid_size.html')
        self.invalid_file = temp_file_with(u'c'*(max_size+1), name=u'invalid_size.html')

    @scenario
    def type_constrained(self):
        self.file_field = FileField(allow_multiple=True, accept=[u'text/*'])
        self.validation_error_message = u'files should be of type text/*'
        self.valid_file = temp_file_with(u'contents', name=u'valid.html')
        self.invalid_file = temp_file_with(u'contents', name=u'invalid.gif')


class MaxNumberOfFilesFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    def new_file_field(self):
        return FileField(allow_multiple=True, max_files=1)


class ToggleableConstraint(ValidationConstraint):
    def __init__(self, fixture=None):
        super(ToggleableConstraint, self).__init__(error_message='test validation message')
        self.fixture = fixture
        
    def validate_input(self, unparsed_input):
        if self.fixture.make_validation_fail:
            raise self

    def __reduce__(self):
        reduced = super(ToggleableConstraint, self).__reduce__()
        pickle_dict = reduced[2]
        del pickle_dict['fixture']
        return reduced

class ToggleValidationFixture(FileUploadInputFixture):
    make_validation_fail = False
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label=u'Attached files')
                fields.files.add_validation_constraint(ToggleableConstraint(fixture))
            @exposed
            def events(self, events):
                events.submit = Event(label=u'Submit')
                                
        return DomainObject()


class StubbedFileUploadInputFixture(FileUploadInputFixture):
    def new_FileUploadForm(self):
        fixture = self
        class FileUploadInputStub(FileUploadInput):
            def create_html_input(self):
                return FileUploadPanelStub(self)
            
        class FileUploadPanelStub(FileUploadPanel):
            def upload_file(self):
                super(FileUploadPanelStub, self).upload_file()
                fixture.file_upload_hook()

        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, u'test')
                self.add_child(FileUploadInputStub(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm

class LargeFileUploadInputFixture(StubbedFileUploadInputFixture):
    def file_upload_hook(self):
        self.simulate_large_file_upload()

    def simulate_large_file_upload(self):
        self.upload_done.wait()

    def simulate_large_file_upload_done(self):
        self.upload_done.set()

    @property
    def web_driver(self):  
        return self.run_fixture.chrome_driver  # These tests only work on chrome

    def new_upload_done(self):
        return threading.Event()


class BrokenFileUploadInputFixture(StubbedFileUploadInputFixture):
    def file_upload_hook(self):
        raise Exception('simulated exception condition')



class FailingConstraint(ValidationConstraint):
    fail = True
    def validate_input(self, unparsed_input):
        if self.fail:
            raise self

@istest
class FileTests(object):
    @test(WebFixture)
    def form_encoding(self, fixture):
        """The enctype of a Form changes to multipart/form-data if it contains an input for a file."""
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.file = FileField(allow_multiple=False, label='Attached files')

        domain_object = DomainObject()

        form = Form(fixture.view, u'testform')
        vassert( u'enctype' not in form.attributes.v )
        
        form.add_child(SimpleFileInput(form, domain_object.fields.file))
        vassert( form.attributes.v[u'enctype'] == u'multipart/form-data' )

    @test(WebFixture)
    def simple_file_input(self, fixture):
        """A SimpleFileInput is a Widget with which a user can choose one or more files.
           The SimpleFileInput transforms the chosen files to UploadedFile objects, and passes these
           to its associated FileField upon a Form submit."""

        file_to_upload = temp_file_with(u'some content')
        class DomainObject(object):
            def __init__(self):
               self.file = None

            @exposed
            def fields(self, fields):
                fields.file = FileField(allow_multiple=False, label=u'Attached files')

            @exposed
            def events(self, events):
                events.upload = Event(label=u'Upload')
                
        domain_object = DomainObject()                
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, u'test')
                self.add_child(SimpleFileInput(self, domain_object.fields.file))
                self.define_event_handler(domain_object.events.upload)
                self.add_child(Button(self, domain_object.events.upload))


        wsgi_app = fixture.new_wsgi_app(child_factory=FileUploadForm.factory(), enable_js=False)
        fixture.reahl_server.set_app(wsgi_app)

        browser = fixture.driver_browser
        browser.open(u'/')

        browser.type(XPath.input_of_type(u'file'), file_to_upload.name)
        browser.click(XPath.button_labelled(u'Upload'))
        vassert( isinstance(domain_object.file, UploadedFile) )
        vassert( domain_object.file.filename == os.path.basename(file_to_upload.name) )
        with domain_object.file.open() as opened_file:
            contents = opened_file.read()
        vassert( contents == u'some content' )


    @test(WebFixture)
    def simple_file_input_exceptions(self, fixture):
        """Usually, when a DomainException happens during a form submit Inputs save the input they received so that
           such input can be pre-populated on the screen rendered by a subsequent GET for a user to correct 
           possible mistakes and resubmit.

           In the case of a SimpleFileInput however, the UploadedFile objects that serve as user input
           to a SimpleFileInput are discarded when a DomainException happens.  This is because it is not
           possible to prepopulate its value using HTML or JS as this would be a security risk.
        """

        file_to_upload = temp_file_with(u'some content')
        failing_constraint = FailingConstraint(u'I am breaking')
        class DomainObject(object):
            def __init__(self):
               self.file = None

            @exposed
            def fields(self, fields):
                fields.file = FileField(allow_multiple=False, label=u'Attached files')
                # FailingConstraint is declared in module level scope for it to be pickleable
                fields.file.add_validation_constraint(failing_constraint)

            @exposed
            def events(self, events):
                events.upload = Event(label=u'Upload')
                
        domain_object = DomainObject()                
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, u'test')
                self.add_child(SimpleFileInput(self, domain_object.fields.file))
                self.define_event_handler(domain_object.events.upload)
                self.add_child(Button(self, domain_object.events.upload))


        wsgi_app = fixture.new_wsgi_app(child_factory=FileUploadForm.factory(), enable_js=False)
        fixture.reahl_server.set_app(wsgi_app)

        browser = fixture.driver_browser
        browser.open(u'/')

        browser.type(XPath.input_of_type(u'file'), file_to_upload.name)
        browser.click(XPath.button_labelled(u'Upload'))
        vassert( browser.is_element_present(u'//label[text()="I am breaking"]') )

        # Message is cleared on second attempt
        failing_constraint.fail = False
        browser.type(XPath.input_of_type(u'file'), file_to_upload.name)
        browser.click(XPath.button_labelled(u'Upload'))
        vassert( not browser.is_element_present(u'//label[text()="I am breaking"]') )


    @test(FileUploadInputFixture)
    def file_upload_input_basics(self, fixture):
        """A FileUploadInput allows its user to upload multiple files one by one before the Form that
           contains the FileUploadInput is submitted.  When the Form is finally submitted
           the FileField of the FileUploadInput receives all the files uploaded as UploadFile objects.
        """
        fixture.reahl_server.set_app(fixture.wsgi_app)

        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.domain_object.submitted )
        vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

        # Upload one file
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))

        vassert( not fixture.domain_object.submitted )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

        # Upload a second file
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        browser.click(XPath.button_labelled(u'Upload'))

        vassert( not fixture.domain_object.submitted )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

        # Submit the form
        browser.click( XPath.button_labelled(u'Submit') )
        vassert( fixture.domain_object.submitted )
        
        # All uploaded files were submitted
        vassert( sorted(fixture.domain_object.submitted_file_info.keys()) == sorted([os.path.basename(f.name) 
                                                                               for f in [fixture.file_to_upload1, fixture.file_to_upload2]] ))

        # Files that were submitted are correct
        file1_content, file1_content_type = fixture.domain_object.submitted_file_info[os.path.basename(fixture.file_to_upload1.name)]
        vassert( file1_content == fixture.file_to_upload1_content )
        vassert( file1_content_type == u'text/html' )



    @test(FileUploadInputFixture)
    def file_upload_input_list_files(self, fixture):
        """The FileUploadInput displays a list of files that were uploaded so far, but is cleared 
           once the Form is submitted."""
        fixture.reahl_server.set_app(fixture.wsgi_app)

        browser = fixture.driver_browser
        browser.open(u'/')

        # Upload one file
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))

        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # Upload a second file
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        browser.click(XPath.button_labelled(u'Upload'))

        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # Submit the form:
        # If an exception is raised, the list is NOT cleared
        fixture.domain_object.throws_exception = True
        browser.click( XPath.button_labelled(u'Submit') )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # Upon successful submit, the list IS cleared
        fixture.domain_object.throws_exception = False
        browser.click( XPath.button_labelled(u'Submit') )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )
        

    @test(FileUploadInputFixture)
    def file_upload_input_remove_files(self, fixture):
        """A user can remove files that were uploaded before the Form which contains the 
           FileUploadInput is submitted."""
        fixture.reahl_server.set_app(fixture.wsgi_app)

        browser = fixture.driver_browser
        browser.open(u'/')

        # Upload two files
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        browser.click(XPath.button_labelled(u'Upload'))

        # Remove file1
        browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload1_name))

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # Only the one file is submitted
        browser.click( XPath.button_labelled(u'Submit') )
        vassert( fixture.domain_object.submitted_file_info.keys() == [fixture.file_to_upload2_name] )
        

    @test(FileUploadInputFixture)
    def file_upload_input_double_uploads(self, fixture):
        """The user is prevented from uploading more than one file with the same name.
        """
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=False))
        browser = fixture.driver_browser
        browser.open(u'/')

        # Upload two files with the same name
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))
        
        # Expect an validation error message
        vassert( browser.is_element_present(u'//label[text()="uploaded files should all have different names"]') )
        vassert( fixture.file_was_uploaded(fixture.file_to_upload1.name) )

    @test(FileUploadInputFixture)
    def async_upload(self, fixture):
        """If JavaScript is enabled, the uploading of files happen in the background via ajax (without reloading the page)
           allowing the user to be busy with the rest of the form. The user does not need to click on the Upload button,
           uploading starts automatically upon choosing a file. The list of uploaded files is appropriately updated.
        """
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        with browser.no_page_load_expected():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)

        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        
        
    @test(LargeFileUploadInputFixture)
    def async_in_progress(self, fixture):
        """While a large file is being uploaded, a progress bar and a Cancel button are displayed. Clicking on the Cancel
           button stops the upload and clears the file name from the list of uploaded files.
        """
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )

        with fixture.reahl_server.in_background():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name) # Upload will block, see fixture

        vassert( browser.is_element_present(u'//ul/li/progress') )
        progress = browser.get_attribute(u'//ul/li/progress', u'value')
        vassert( progress == u'100' )
        browser.click(XPath.button_labelled(u'Cancel'))

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
            

    @test(FileUploadInputFixture)
    def prevent_duplicate_upload_js(self, fixture):
        """The user is prevented from uploading more than one file with the same name on the client side.
        """
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        vassert( not browser.is_element_present(u'//label[text()="uploaded files should all have different names"]') )

        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        vassert( not browser.is_element_present(u'//label[text()="uploaded files should all have different names"]') )

        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        vassert( browser.is_element_present(u'//label[text()="uploaded files should all have different names"]') )

        browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload2_name))
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        vassert( not browser.is_element_present(u'//label[text()="uploaded files should all have different names"]') )

        
    @test(LargeFileUploadInputFixture)
    def prevent_form_submit(self, fixture):
        """The user is prevented from submitting the Form while one or more file uploads are still in progress."""
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        with fixture.reahl_server.in_background():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)

        with browser.no_page_load_expected():
            browser.click( XPath.button_labelled(u'Submit'), wait=False )

            alert = fixture.web_driver.switch_to_alert()
            vassert( alert.text == u'Please try again when all files have finished uploading.' )
            alert.accept()
        
        
    @test(FileUploadInputFixture)
    def async_remove(self, fixture):
        """With javascript enabled, removing of uploaded files take place via ajax."""
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        # Upload two files
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        browser.click(XPath.button_labelled(u'Upload'))
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        browser.click(XPath.button_labelled(u'Upload'))

        # Remove file1
        with browser.no_page_load_expected():
            browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload1_name))

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # The javascript works on DOM elements that have been generated server-side as well:
        browser.refresh()
        with browser.no_page_load_expected():
            browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload2_name))

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        # No files are submitted eventually
        browser.click( XPath.button_labelled(u'Submit') )
        vassert( fixture.domain_object.submitted_file_info.keys() == [] )

    @test(BrokenFileUploadInputFixture)
    def async_upload_error(self, fixture):
        """If an error happens during (ajax) upload, the user is notified."""
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
        fixture.config.reahlsystem.debug = False # So that we don't see the exception output while testing
        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not browser.is_element_present(XPath.label_with_text('an error ocurred, please try again later.')) )

        with expected(Exception):
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)

        vassert( browser.is_element_present(XPath.label_with_text('an error occurred, please try again later.')) )
        vassert( not browser.is_element_enabled(XPath.button_labelled(u'Cancel')) )

    @test(ToggleValidationFixture)
    def async_upload_domain_exception(self, fixture):
        """When a DomainException happens upon uploading via JavaScript, 
           the form is replaced with a rerendered version from the server."""
        
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
        browser = fixture.driver_browser
        browser.open(u'/')

        fixture.make_validation_fail = False
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)

        fixture.make_validation_fail = True
        browser.execute_script(u'$(".reahl-upload-panel").children().addClass("reload_flag")')
        with browser.no_page_load_expected():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        has_class = browser.execute_script(u'$(".reahl-upload-panel").children().hasClass("reload_flag")')
        vassert( not has_class ) # ie, the UploadPanel has been reloaded

        # JS Stuff on re-rendered form still work
        
        # 1: Server-rendered validation message has been cleared
        vassert( browser.is_element_present(XPath.label_with_text('test validation message')) )
        fixture.make_validation_fail = False
        with browser.no_page_load_expected():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        vassert( not browser.is_element_present(XPath.label_with_text('test validation message')) )

        # 2: The remove button still happens via ajax
        with browser.no_page_load_expected():
            browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload1_name))
            browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload2_name))

    @test(LargeFileUploadInputFixture)
    def queueing_async_uploads(self, fixture):
        """Asynchronous uploads do not happen concurrently, they are queued one after another.
        """
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )

        with fixture.reahl_server.in_background():
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name) # Upload will block, see fixture
            browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name) # Upload will block, see fixture

        progress1 = browser.get_attribute(u'//ul/li[1]/progress', u'value')
        progress2 = browser.get_attribute(u'//ul/li[2]/progress', u'value')
        vassert( progress1 == u'100' )
        vassert( progress2 == u'0' )

        fixture.simulate_large_file_upload_done()

        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
        vassert( fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

    @test(PerFileConstrainedFileUploadInputFixture)
    def async_validation(self, fixture):
        """Validations are checked in JavaScript before uploading.
        """
        # Only tested for the FileUploadInput, as it uses the SimpleFileInput
        # in its own implementation, in a NestedForm, and has to pass on the
        # filesize constraint all the way. This way, we test all of that.
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.uploaded_file_is_listed( fixture.valid_file.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) )

        browser.type(XPath.input_of_type(u'file'), fixture.invalid_file.name)
        vassert( not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) )
        vassert( browser.is_element_present(XPath.label_with_text(fixture.validation_error_message)) )

        browser.type(XPath.input_of_type(u'file'), fixture.valid_file.name)
        vassert( fixture.uploaded_file_is_listed( fixture.valid_file.name ) )
    
    @test(MaxNumberOfFilesFileUploadInputFixture)
    def async_number_files_validation(self, fixture):
        """A Field set to only allow a maximum number of files is checked for validity before uploading in JS.
        """
        # Only tested for the FileUploadInput, as it uses the SimpleFileInput
        # in its own implementation, in a NestedForm, and has to pass on the
        # filesize constraint all the way. This way, we test all of that.
        fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
        
        browser = fixture.driver_browser
        browser.open(u'/')

        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload1.name)
        vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )

        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )
        vassert( browser.is_element_present(XPath.label_with_text('a maximum of 1 files may be uploaded')) )

        browser.click(XPath.button_labelled(u'Remove', filename=fixture.file_to_upload1_name))
        browser.type(XPath.input_of_type(u'file'), fixture.file_to_upload2.name)
        vassert( not browser.is_element_present(XPath.label_with_text('a maximum of 1 files may be uploaded')) )




# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import contextlib
import pkg_resources

from webob import Request, Response

from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected
from reahl.stubble import EmptyStub, stubclass

from reahl.component.config import Configuration, ReahlSystemConfig
from reahl.component.exceptions import IncorrectArgumentError, IsInstance
from reahl.web.egg import WebConfig, ReahlEgg
from reahl.webelixirimpl import WebUserSession, PersistedException, UserInput
from reahl.web.fw import UrlBoundView, UserInterface, WebExecutionContext, Widget
from reahl.web.ui import TwoColumnPage, P, Div, Slot
from reahl.webdev.tools import WidgetTester, Browser
from reahl.web_dev.fixtures import WebBasicsMixin, WebFixture


class WidgetFixture(Fixture, WebBasicsMixin):
    def new_user_interface_factory(self):
        factory = UserInterface.factory(u'test_user_interface_name')
        factory.attach_to(u'/', {})
        return factory
        
    def new_user_interface(self):
        factory = self.ui_factory
        with self.context:
            return factory.create()

    def new_view(self, user_interface=None, relative_path=u'/', title='A view', slot_definitions={}):
        user_interface = user_interface or self.user_interface
        return UrlBoundView(user_interface, relative_path, title, slot_definitions)


@istest
class WidgetBasics(object):
    @test(WebFixture)
    def basic_widget(self, fixture):
        """Basic widgets are created by adding children widgets to them, and the result is rendered in HTML"""
        
        class MyMessage(Div):
            def __init__(self, view):
                super(MyMessage, self).__init__(view)
                self.add_child(P(view, text=u'Hello World!'))
                self.add_children([P(view, text=u'a'), P(view, text=u'b')])

        message = MyMessage(fixture.view)
        
        widget_tester = WidgetTester(message)
        actual = widget_tester.render_html()
        
        vassert( actual == u'<div><p>Hello World!</p><p>a</p><p>b</p></div>' )

    @test(WebFixture)
    def visibility(self, fixture):
        """Widgets are rendered only if their .visible property is True."""
        
        class MyMessage(Widget):
            is_visible = True
            def __init__(self, view):
                super(MyMessage, self).__init__(view)
                self.add_child(P(view, text=u'你好世界!'))
                
            @property
            def visible(self):
                return self.is_visible
 
        message = MyMessage(fixture.view)
        
        widget_tester = WidgetTester(message)

        # Case: when visible
        message.is_visible = True
        actual = widget_tester.render_html()
        vassert( actual == u'<p>你好世界!</p>' )

        # Case: when not visible
        message.is_visible = False
        actual = widget_tester.render_html()
        vassert( actual == u'' )

    @test(WebFixture)
    def widget_factories_and_args(self, fixture):
        """Widgets can be created from factories which allow you to supply widget-specific args 
           and/or kwargs before a view is available."""
        
        class WidgetWithArgs(Widget):
            def __init__(self, view, arg, kwarg=None):
                super(WidgetWithArgs, self).__init__(view)
                self.saved_arg = arg
                self.saved_kwarg = kwarg

        factory = WidgetWithArgs.factory('a', kwarg='b')

        widget = factory.create(fixture.view)
        
        vassert( widget.saved_arg == 'a' )
        vassert( widget.saved_kwarg == 'b' )

    @test(WebFixture)
    def widget_factories_error(self, fixture):
        """Supplying arguments to .factory that do not match those of the Widget's __init__ is reported to the programmer.."""
        
        class WidgetWithArgs(Widget):
            def __init__(self, view, arg, kwarg=None):
                super(WidgetWithArgs, self).__init__(view)
                self.saved_arg = arg
                self.saved_kwarg = kwarg

        def check_exc(ex):
            vassert( str(ex).startswith("An attempt was made to create a WidgetFactory for <class 'reahl.web_dev.widgets.widgetbasics.WidgetWithArgs'> with arguments that do not match what is expected for <class 'reahl.web_dev.widgets.widgetbasics.WidgetWithArgs'>") )
        with expected(IncorrectArgumentError, test=check_exc):
            factory = WidgetWithArgs.factory('a', 'b', 'c')

    @test(WebFixture)
    def widget_construct_error(self, fixture):
        """Passing anything other than a View as a Widget's view argument on construction results in an error."""

        with expected(IsInstance):
            Widget(EmptyStub())

    @test(WebFixture)
    def widget_adding_error(self, fixture):
        """Passing anything other than other Widgets to .add_child or add_children results in an error."""

        widget = Widget(fixture.view)

        with expected(IsInstance):
            widget.add_child(EmptyStub())

        with expected(IsInstance):
            widget.add_children([Widget(fixture.view), EmptyStub()])

    @test(WebFixture)
    def basic_working_of_slots(self, fixture):
        """Slots are special Widgets that can be added to the page. The contents of a
           Slot are then supplied (differently) by different Views."""

        class MyPage(Widget):
            def __init__(self, view):
                super(MyPage, self).__init__(view)
                self.add_child(Slot(view, u'slot1'))
                self.add_child(Slot(view, u'slot2'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(MyPage)

                home = self.define_view(u'/', title=u'Home')
                home.set_slot(u'slot1', P.factory(text=u'a'))
                home.set_slot(u'slot2', P.factory(text=u'b'))

                other = self.define_view(u'/other', title=u'Other')
                other.set_slot(u'slot1', P.factory(text=u'other'))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        
        browser.open('/')
        [slot1_p, slot2_p] = browser.lxml_html.xpath('//p')
        vassert( slot1_p.text == u'a' )
        vassert( slot2_p.text == u'b' )

        browser.open('/other')
        [slot1_p] = browser.lxml_html.xpath('//p')
        vassert( slot1_p.text == u'other' )


    @test(WebFixture)
    def defaults_for_slots(self, fixture):
        """A Widget can have defaults for its slots."""

        class MyPage(Widget):
            def __init__(self, view):
                super(MyPage, self).__init__(view)
                self.add_child(Slot(view, u'slot3'))
                self.add_default_slot(u'slot3', P.factory(text=u'default'))
               
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(MyPage)
                self.define_view(u'/', title=u'Home')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        
        browser.open('/')
        [slot3_p] = browser.lxml_html.xpath('//p')
        vassert( slot3_p.text == u'default' )


    @test(WebFixture)
    def activating_javascript(self, fixture):
        """The JavaScript snippets of all Widgets are collected in a jQuery.ready() function by"""
        """ an automatically inserted Widget in the slot named reahl_header.  Duplicate snippets are removed."""
        @stubclass(Widget)
        class WidgetWithJavaScript(Widget):
            def __init__(self, view, fake_js):
                super(WidgetWithJavaScript, self).__init__(view)
                self.fake_js = fake_js
                
            def get_js(self, context=None):
                return [self.fake_js]

        class MyPage(Widget):
            def __init__(self, view):
                super(MyPage, self).__init__(view)
                self.add_child(Slot(view, u'reahl_header'))
                self.add_child(WidgetWithJavaScript(view, u'js1'))
                self.add_child(WidgetWithJavaScript(view, u'js2'))
                self.add_child(WidgetWithJavaScript(view, u'js1'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(MyPage)
                self.define_view(u'/', title=u'Home')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        
        browser.open('/')
        rendered_js = browser.lxml_html.xpath('//script')[1].text
        vassert( rendered_js == u'\njQuery(document).ready(function($){\n$(\'body\').addClass(\'enhanced\');\njs1\njs2\n\n});\n' )
        
        number_of_duplicates = rendered_js.count('js1') - 1
        vassert( number_of_duplicates == 0 )


    class AttachmentScenarios(WebFixture):
        @scenario
        def javascript(self):
            self.static_file = u'/static/reahl.js'
            self.attachment_label = u'js'

        @scenario
        def css(self):
            self.static_file = u'/static/reahl.css'
            self.attachment_label = u'css'

    @test(AttachmentScenarios)
    def shipping_attachments(self, fixture):
        """The JavaScript and CSS files listed in the .reahlproject are discovered when the webserver starts up, and put into one file
           for inclusion on each page of a Reahl web application."""

        wsgi_app = fixture.new_wsgi_app(enable_js=True)
        browser = Browser(wsgi_app)
        browser.open(fixture.static_file)

        def broken_but_comparable_minify(some_js):
            minified = re.sub(u'/\*.*\*/', u'', some_js)
            minified = re.sub(u'//.*\n', u'', minified)
            minified = re.sub(u'\n', u'', minified)
            minified = re.sub(u'\t', u'', minified)
            minified = re.sub(u' ', u'', minified)
            return minified

        import reahl.web
        reahl_egg = ReahlEgg(pkg_resources.require('reahl-web')[0])
        attachments = reahl_egg.find_attachments(fixture.attachment_label)

        # Only the js/css of one widget is checked to check the mechanism...
        with open(attachments[0].filename, u'r') as snippet_file:
            snippet = broken_but_comparable_minify(snippet_file.read())

        served_statics = broken_but_comparable_minify(browser.raw_html)

        vassert( served_statics.find(snippet) > 0 )











# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from nose.tools import istest
from reahl.tofu import Fixture, test, vassert, expected
from reahl.component.exceptions import ProgrammerError

from reahl.webdev.tools import WidgetTester, Browser
from reahl.web_dev.fixtures import WebFixture, WebBasicsMixin, ReahlWSGIApplicationStub


from reahl.component.modelinterface import Field, exposed, IntegerField
from reahl.web.fw import Widget, Bookmark, ReahlWSGIApplication
from reahl.web.ui import A, P, Form, TextInput, Panel



class QueryStringFixture(Fixture, WebBasicsMixin):
    def is_state_now(self, state):
        return self.driver_browser.execute_script('return (window.jQuery("p").html() == "My state is now %s")' % state)

    def change_fragment(self, fragment):
        self.driver_browser.execute_script(u'return (window.location.hash="%s")' % fragment)

    def new_FancyWidget(self):
        fixture = self
        class MyFancyWidget(Panel):
            def __init__(self, view):
                super(MyFancyWidget, self).__init__(view, css_id=u'sedrick')
                self.enable_refresh()
                self.add_child(P(self.view, text=u'My state is now %s' % self.fancy_state))
                fixture.widget = self

            @exposed
            def query_fields(self, fields):
                fields.fancy_state = IntegerField(required=False, default=1)
            

        return MyFancyWidget

    def new_wsgi_app(self, widget_factory=None):
        widget_factory = widget_factory or self.FancyWidget.factory()
        return super(QueryStringFixture, self).new_wsgi_app(enable_js=True, 
                                                          child_factory=widget_factory)



@istest
class WidgetQueryArgTests(object):
    @test(QueryStringFixture)
    def query_string_widget_arguments(self, fixture):
        """Widgets can have arguments that are read from a query string"""

        class WidgetWithQueryArguments(Widget):
            def __init__(self, view):
                super(WidgetWithQueryArguments, self).__init__(view)
                self.add_child(P(view, text=self.arg_directly_on_widget))

            @exposed
            def query_fields(self, fields):
                fields.arg_directly_on_widget = Field()

        wsgi_app = fixture.new_wsgi_app(widget_factory=WidgetWithQueryArguments.factory())
        browser = Browser(wsgi_app)

        browser.open(u'/?arg_directly_on_widget=supercalafragalisticxpelidocious')
        vassert( browser.lxml_html.xpath(u'//p')[0].text == u'supercalafragalisticxpelidocious' )


    @test(QueryStringFixture)
    def query_string_prepopulates_form(self, fixture):
        """Widget query string arguments can be used on forms to pre-populate inputs based on the query string."""

        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.arg_on_other_object = Field()

        class FormWithQueryArguments(Form):
            def __init__(self, view):
                self.model_object = ModelObject()
                super(FormWithQueryArguments, self).__init__(view, u'name')
                self.add_child(TextInput(self, self.model_object.fields.arg_on_other_object))

            @exposed
            def query_fields(self, fields):
                fields.arg_on_other_object = self.model_object.fields.arg_on_other_object
        

        wsgi_app = fixture.new_wsgi_app(widget_factory=FormWithQueryArguments.factory())
        browser = Browser(wsgi_app)

        browser.open(u'/?arg_on_other_object=metoo')
        vassert( browser.lxml_html.xpath(u'//input')[0].value == u'metoo' )


    @test(QueryStringFixture)
    def widgets_with_bookmarkable_state(self, fixture):
        """If a widget has query_fields, call `.enable_refresh()` on it to let it change 
           its contents without reloading the whole page,
           depending on a variable in the fragment of the current URL. This makes
           such Widgets bookmarkable by a browser."""

        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        # Case: the default
        vassert( fixture.driver_browser.wait_for(fixture.is_state_now, 1) )
        vassert( fixture.widget.fancy_state == 1 )

        # Case: change without page load
        fixture.change_fragment(u'#fancy_state=2')
        vassert( fixture.driver_browser.wait_for(fixture.is_state_now, 2) )
        vassert( fixture.widget.fancy_state == 2 )

        # Case: unrelated fragment changes do not trigger a reload of the widget
        previous_widget = fixture.widget
        fixture.change_fragment(u'#fancy_state=2&other_var=other_value')
        vassert( fixture.driver_browser.wait_for(fixture.is_state_now, 2) )
        vassert( fixture.widget is previous_widget )

    @test(WebFixture)
    def css_id_is_mandatory(self, fixture):
        """If a Widget is enabled to to be refreshed, it must also have a css_id set."""
        class MyFancyWidget(Panel):
            def __init__(self, view):
                super(MyFancyWidget, self).__init__(view)
                self.enable_refresh()

        with expected(ProgrammerError):
            MyFancyWidget(fixture.view)


    @test(QueryStringFixture)
    def bookmarks_support_such_fragments(self, fixture):
        """Page-internal bookmarks support such bookmarkable widgets.

           These Bookmarks usually do not affect an URL - they just set widget states in different ways,
           depending on whether the client has javascript support or not.
           However, if a page was opened using the widget arguments on the querystring, a bookmark that would
           normally only have changed that argument on the hash will point to a new
           url on which the argument has been removed from the querystring and changed on the hash.
        """

        internal_bookmark = Bookmark(u'', u'', u'an ajax bookmark', query_arguments={u'fancy_state':2}, ajax=True)
#        internal_bookmark = fixture.MyFancyWidget.get_bookmark(fancy_state=2)
        normal_bookmark = Bookmark(u'/', u'', u'a normal bookmark')

        # You can query whether a bookmark is page_internal or not
        vassert( internal_bookmark.is_page_internal )
        vassert( not normal_bookmark.is_page_internal )

        # page-internal bookmarks must be added to normal ones to be useful
        usable_bookmark = normal_bookmark+internal_bookmark
        
        wsgi_app = fixture.new_wsgi_app(widget_factory=A.factory_from_bookmark(usable_bookmark))

        # Case: when rendered without javascript
        browser = Browser(wsgi_app)
        browser.open(u'/')

        a = browser.lxml_html.xpath(u'//a')[0]
        vassert( a.attrib[u'href'] == u'/?fancy_state=2' )
        vassert( a.text == u'an ajax bookmark' )

        # Case: when rendered in a browser with javascript support
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')
        vassert(     fixture.driver_browser.is_element_present(u"//a[@href='/#fancy_state=2']") )
        vassert( not fixture.driver_browser.is_element_present(u"//a[@href='/?fancy_state=2']") )

        # Case: when the argument was given on the query string of the current page
        fixture.driver_browser.open('/?fancy_state=4')
        vassert(     fixture.driver_browser.is_element_present(u"//a[@href='/#fancy_state=2']") )





from setuptools import setup
setup(
    name='reahl-web',
    version=u'2.1.1',
    description='The core Reahl web framework',
    long_description=u'Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nThis package contains the core of the Reahl framework.\n\nSee http://www.reahl.org/docs/current/tutorial/gettingstarted.d.html for installation instructions. ',
    url=u'http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.messages', 'reahl.web', 'reahl.web_dev', 'reahl.web.static', 'reahl.web_dev.inputandvalidation', 'reahl.web_dev.widgets', 'reahl.web_dev.appstructure', 'reahl.web_dev.advanced', 'reahl.web.static.jquery', 'reahl.web_dev.advanced.subresources'],
    py_modules=['setup'],
    include_package_data=True,
    package_data={'': [u'*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl', 'reahl.messages'],
    install_requires=[u'reahl-component>=2.1,<2.2', u'reahl-interfaces>=2.1,<2.2', u'slimit>=0.8,<0.9', u'cssmin>=0.1,<0.2', u'BeautifulSoup>=3.2,<3.3', u'webob>=1.2,<1.3', u'Babel>=0.9,<0.10'],
    setup_requires=[],
    tests_require=[u'reahl-tofu>=2.1,<2.2', u'reahl-stubble>=2.1,<2.2', u'reahl-sqlalchemysupport>=2.1,<2.2', u'reahl-web-elixirimpl>=2.1,<2.2', u'reahl-domain>=2.1,<2.2', u'reahl-dev>=2.1,<2.2', u'reahl-webdev>=2.1,<2.2'],
    test_suite=u'reahl.web_dev',
    entry_points={
        'reahl.translations': [
            u'reahl-web = reahl.messages'    ],
        u'reahl.configspec': [
            u'config = reahl.web.egg:WebConfig'    ],
        u'reahl.attachments.css': [
            u'0:reahl/web/reahl.labelledinput.css = reahl',
            u'1:reahl/web/reahl.labeloverinput.css = reahl',
            u'2:reahl/web/reahl.menu.css = reahl',
            u'3:reahl/web/reahl.hmenu.css = reahl',
            u'4:reahl/web/reahl.slidingpanel.css = reahl',
            u'5:reahl/web/reahl.runningonbadge.css = reahl'    ],
        u'reahl.attachments.any': [
            u'0:reahl/web/static/runningon.png = reahl'    ],
        u'reahl.attachments.js': [
            u'0:reahl/web/reahl.hashchange.js = reahl',
            u'1:reahl/web/reahl.ajaxlink.js = reahl',
            u'2:reahl/web/reahl.form.js = reahl',
            u'3:reahl/web/reahl.textinput.js = reahl',
            u'4:reahl/web/reahl.cueinput.js = reahl',
            u'5:reahl/web/reahl.labeloverinput.js = reahl',
            u'6:reahl/web/reahl.fileuploadli.js = reahl',
            u'7:reahl/web/reahl.fileuploadpanel.js = reahl',
            u'8:reahl/web/reahl.popupa.js = reahl',
            u'9:reahl/web/reahl.slidingpanel.js = reahl'    ],
        'reahl.eggs': [
            u'Egg = reahl.component.eggs:ReahlEgg'    ],
        u'reahl.persistlist': [
            u'0 = reahl.web.pager:SequentialPageIndex'    ],
                 },
    extras_require={}
)



# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from webob.exc import HTTPNotFound
from webob import Request

from nose.tools import istest
from reahl.tofu import Fixture, test
from reahl.tofu import vassert, expected, NoException
from reahl.stubble import EmptyStub

from reahl.web_dev.fixtures import WebFixture, ReahlWSGIApplicationStub
from reahl.webdev.tools import Browser, XPath
from reahl.component.modelinterface import Event, Field, Action, exposed, IntegerField
from reahl.component.exceptions import ProgrammerError
from reahl.web.ui import Form, TwoColumnPage, Button, A
from reahl.web.fw import UserInterface, ViewPreCondition, Redirect, Detour, Return, IdentityDictionary, UrlBoundView
from reahl.component.context import ExecutionContext


class FormWithButton(Form):
    def __init__(self, view, event):
        super(FormWithButton, self).__init__(view, u'test_events')
        self.add_child(Button(self, event))
                
@istest
class ControlledUserInterfacesTests(object):
    @test(WebFixture)
    def basic_transition(self, fixture):
        """Transitions express how the browser is ferried between Views in reaction to user-initiated Events."""
        def do_something():
            fixture.did_something = True

        class UIWithTwoViews(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me', action=Action(do_something))
                event.bind(u'anevent', None)
                slot_definitions = {u'main': FormWithButton.factory(event)}
                viewa = self.define_view(u'/viewa', title=u'View a', slot_definitions=slot_definitions)
                viewb = self.define_view(u'/viewb', title=u'View b', slot_definitions=slot_definitions)
                self.define_transition(event, viewa, viewb)

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithTwoViews,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # The transition works from viewa
        fixture.did_something = False
        browser.open('/a_ui/viewa')
        browser.click('//input[@value="Click me"]')
        vassert( browser.location_path == u'/a_ui/viewb' )
        vassert( fixture.did_something )

        # The transition does not work from viewb
        fixture.did_something = False
        browser.open('/a_ui/viewb')
        with expected(ProgrammerError):
            browser.click('//input[@value="Click me"]')
        vassert( not fixture.did_something )

    @test(WebFixture)
    def guards(self, fixture):
        """Guards can be set on Transitions. A Transition is only elegible for firing if its guard is True."""
        fixture.guard_value = None
        adjustable_guard = Action(lambda:fixture.guard_value)

        false_guard = Action(lambda:False)

        class UIWithGuardedTransitions(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me')
                event.bind(u'anevent', None)
                slot_definitions = {u'main': FormWithButton.factory(event)}
                viewa = self.define_view(u'/viewa', title=u'View a', slot_definitions=slot_definitions)
                viewb = self.define_view(u'/viewb', title=u'View b')
                viewc = self.define_view(u'/viewc', title=u'View c')
                self.define_transition(event, viewa, viewb, guard=false_guard)
                self.define_transition(event, viewa, viewc, guard=adjustable_guard)

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithGuardedTransitions,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # The transition with True guard is the one followed
        fixture.guard_value = True
        browser.open('/a_ui/viewa')
        browser.click('//input[@value="Click me"]')
        vassert( browser.location_path == u'/a_ui/viewc' )

        # If there is no Transition with a True guard, fail
        fixture.guard_value = False
        browser.open('/a_ui/viewa')
        with expected(ProgrammerError):
            browser.click('//input[@value="Click me"]')

    @test(WebFixture)
    def local_transition(self, fixture):
        """A local Transition has its source as its target."""
        def do_something():
            fixture.did_something = True

        fixture.guard_passes = True
        guard = Action(lambda:fixture.guard_passes)
        
        class UIWithAView(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me', action=Action(do_something))
                event.bind(u'anevent', None)
                slot_definitions = {u'main': FormWithButton.factory(event)}
                viewa = self.define_view(u'/viewa', title=u'View a', slot_definitions=slot_definitions)
                self.define_local_transition(event, viewa, guard=guard)

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithAView,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # The transition works from viewa
        fixture.did_something = False
        browser.open('/a_ui/viewa')
        browser.click('//input[@value="Click me"]')
        vassert( browser.location_path == u'/a_ui/viewa' )
        vassert( fixture.did_something )

        # But it is also guarded
        fixture.guard_passes = False
        browser.open('/a_ui/viewa')
        with expected(ProgrammerError):
            browser.click('//input[@value="Click me"]')


    @test(WebFixture)
    def transitions_to_parameterised_views(self, fixture):
        """When a Button is placed for an Event that may trigger a Transition to a parameterised View,
           the Event should be bound to the arguments to be used for the target View, using .with_arguments()"""

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', event_argument1=IntegerField(), 
                                        event_argument2=IntegerField())

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                event = model_object.events.an_event.with_arguments(event_argument1=1, event_argument2=2, view_argument=3)
                self.add_child(Button(self, event))

        class ParameterisedView(UrlBoundView):
            def assemble(self, event_argument1=None, view_argument=None):
                self.title = u'View with event_argument1: %s%s and view_argument: %s%s' \
                            % (event_argument1, type(event_argument1), view_argument, type(view_argument))
                
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')

                other_view = self.define_view(u'/page2', title=u'Page 2', 
                                              view_class=ParameterisedView, 
                                              event_argument1=IntegerField(required=True),
                                              view_argument=IntegerField(required=True))
                home.set_slot(u'main', MyForm.factory(u'myform'))

                self.define_transition(model_object.events.an_event, home, other_view)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # when the Action is executed, the correct arguments are passed to the View
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( fixture.driver_browser.title == u'View with event_argument1: 1<type \'int\'> and view_argument: 3<type \'int\'>' )


    @test(WebFixture)
    def transitions_to_parameterised_views_error(self, fixture):
        """If an Event triggers a Transition to a parameterised View, and it was not bound to the arguments
           expected by the target View, an error is raised."""

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'Click me')

        model_object = ModelObject()

        class FormWithIncorrectButtonToParameterisedView(Form):
            def __init__(self, view):
                super(FormWithIncorrectButtonToParameterisedView, self).__init__(view, u'test_events')
                self.add_child(Button(self, model_object.events.an_event.with_arguments(arg1='1', arg2='2')))

        class ParameterisedView(UrlBoundView):
            def assemble(self, object_key=None):
                self.title = u'View for: %s' % object_key

        class UIWithParameterisedViews(UserInterface):
            def assemble(self):
                slot_definitions = {u'main': FormWithIncorrectButtonToParameterisedView.factory()}
                normal_view = self.define_view(u'/static', title=u'Static', slot_definitions=slot_definitions)
                parameterised_view = self.define_view(u'/dynamic', view_class=ParameterisedView, object_key=Field(required=True))
                self.define_transition(model_object.events.an_event, normal_view, parameterised_view)

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithParameterisedViews,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/static')
        with expected(ProgrammerError):
            browser.click(XPath.button_labelled(u'Click me'))


    @test(WebFixture)
    def view_preconditions(self, fixture):
        """Views can have Conditions attached to them - if a ViewPreCondition fails upon a GET or HEAD request,
           the specified exception is raised. For all other requests, HTTPNotFound is raised."""
        class SomeException(Exception):
            pass
            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                slot_definitions = {u'main': Form.factory(u'the_form')}
                view = self.define_view(u'/', title=u'Hello', slot_definitions=slot_definitions)
                failing_precondition = ViewPreCondition(lambda: False, exception=SomeException)
                passing_precondition = ViewPreCondition(lambda: True)
                view.add_precondition(passing_precondition)
                view.add_precondition(failing_precondition)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        with expected(SomeException):
            browser.open('/')

        browser.post('/_the_form', {}, status=404)
      
    @test(WebFixture)
    def inverse_view_preconditions(self, fixture):
        """A ViewPreCondition can give you another ViewPreCondition which it itself negated, optionally with its own exception."""
        class SomeException(Exception):
            pass
            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                view = self.define_view(u'/', title=u'Hello')
                passing_precondition = ViewPreCondition(lambda: True)
                failing_precondition = passing_precondition.negated(exception=SomeException)
                view.add_precondition(failing_precondition)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        with expected(SomeException):
            browser.open('/')
      
    @test(WebFixture)
    def redirect(self, fixture):
        """Redirect is a special exception that will redirect the browser to another View."""
        class UIWithRedirect(UserInterface):
            def assemble(self):
                viewa = self.define_view(u'/viewa', title=u'A')
                viewb = self.define_view(u'/viewb', title=u'B')
                failing_precondition = ViewPreCondition(lambda: False, exception=Redirect(viewb.as_bookmark(self)))
                viewa.add_precondition(failing_precondition)
            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithRedirect,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/viewa')
        vassert( browser.location_path == u'/a_ui/viewb' )

    @test(WebFixture)
    def detours_and_return_transitions(self, fixture):
        """Detour is a special exception that will redirect the browser to another View, but it also does the
           necessary housekeeping that will allow a return_transition to let the browser return to where the 
           Detour was thrown."""
        
        fixture.make_precondition_pass = False
        class UIWithDetour(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me')
                event.bind(u'anevent', None)
                viewa = self.define_view(u'/viewa', title=u'View a')
                slot_with_button = {u'main': FormWithButton.factory(event)}
                
                step1 = self.define_view(u'/firstStepOfDetour', title=u'Step 1', slot_definitions=slot_with_button)
                step2 = self.define_view(u'/lastStepOfDetour', title=u'Step 2', slot_definitions=slot_with_button)

                viewa.add_precondition(ViewPreCondition(lambda: fixture.make_precondition_pass, exception=Detour(step1.as_bookmark(self))))
                self.define_transition(event, step1, step2)
                self.define_return_transition(event, step2)
            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithDetour,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        fixture.did_something = False

        fixture.make_precondition_pass = False
        browser.open(u'/a_ui/viewa')
        vassert( browser.location_path == u'/a_ui/firstStepOfDetour' )
        
        browser.click(u'//input[@type="submit"]')
        vassert( browser.location_path == u'/a_ui/lastStepOfDetour' )
                
        fixture.make_precondition_pass = True
        browser.click(u'//input[@type="submit"]')
        vassert( browser.location_path == u'/a_ui/viewa' )

        # The query string is cleared after such a return (it is used to remember where to return to)
        vassert( browser.location_query_string == u'' )
        
        
    @test(WebFixture)
    def detours_and_explicit_return_view(self, fixture):
        """A Detour can also explicitly set the View to return to."""
        
        class UIWithDetour(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me')
                event.bind(u'anevent', None)
                viewa = self.define_view(u'/viewa', title=u'View a')
                explicit_return_view = self.define_view(u'/explicitReturnView', title=u'Explicit Return View')

                slot_with_button = {u'main': FormWithButton.factory(event)}
                detour = self.define_view(u'/detour', title=u'Detour', slot_definitions=slot_with_button)

                viewa.add_precondition(ViewPreCondition(lambda: False, exception=Detour(detour.as_bookmark(self), return_to=explicit_return_view.as_bookmark(self))))
                self.define_return_transition(event, detour)
            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithDetour,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        
        browser.open(u'/a_ui/viewa')
        vassert( browser.location_path == u'/a_ui/detour' )
        
        browser.click(u'//input[@type="submit"]')
        vassert( browser.location_path == u'/a_ui/explicitReturnView' )

        # The query string is cleared after such a return (it is used to remember where to return to)
        vassert( browser.location_query_string == u'' )
                
        
    @test(WebFixture)
    def redirect_used_to_return(self, fixture):
        """A Return is an exception used with Preconditoins to return automatically to another View (as set by detour),
           instead of using a return_transition (the latter can only be triggered by a user)."""
        
        class UIWithDetour(UserInterface):
            def assemble(self):
                viewa = self.define_view(u'/viewa', title=u'View a')
                explicit_return_view = self.define_view(u'/explicitReturnView', title=u'Explicit Return View')
                default = self.define_view(u'/defaultReturnView', title=u'Default view to return to')
                detour = self.define_view(u'/detour', title=u'Detour')

                viewa.add_precondition(ViewPreCondition(lambda: False, exception=Detour(detour.as_bookmark(self), return_to=explicit_return_view.as_bookmark(self))))
                detour.add_precondition(ViewPreCondition(lambda: False, exception=Return(default.as_bookmark(self))))

            
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithDetour,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # Normal operation - when a caller can be determined
        browser.open(u'/a_ui/viewa')
        vassert( browser.location_path == u'/a_ui/explicitReturnView' )

        #  - the query string is cleared after such a return (it is used to remember where to return to)
        vassert( browser.location_query_string == u'' )
        
        # When a caller cannot be determined, the default is used
        browser.open(u'/a_ui/detour')
        vassert( browser.location_path == u'/a_ui/defaultReturnView' )
        
        #  - the query string is cleared after such a return (it is used to remember where to return to)
        vassert( browser.location_query_string == u'' )


    @test(WebFixture)
    def unconditional_redirection(self, fixture):
        """You can force an URL to always redirect to a given Bookmark."""

        class UIWithRedirect(UserInterface):
            def assemble(self):
                self.define_view(u'/target', title=u'')
                self.define_redirect(u'/redirected', self.get_bookmark(relative_path=u'/target'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithRedirect,  IdentityDictionary(), name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/redirected')
        vassert( browser.location_path == u'/a_ui/target' )

    @test(WebFixture)
    def linking_to_views_marked_as_detour(self, fixture):
        """A View can be marked as the start of a Detour. Where used, a Bookmark for such a View
           will automatically include a returnTo in the its query string. This allows an
           eventual return_transition (or similar) to return to where, eg, a link was clicked from.
           This mechanism works for returning across UserInterfaces."""

        class UIWithLink(UserInterface):
            def assemble(self, bookmark=None):
                self.bookmark = bookmark
                slot_definitions = {u'main': A.factory_from_bookmark(self.bookmark)}
                self.define_view(u'/initial', title=u'View a', slot_definitions=slot_definitions)

        class UIWithDetour(UserInterface):
            def assemble(self):
                event = Event(label=u'Click me')
                event.bind(u'anevent', None)
                slot_with_button = {u'main': FormWithButton.factory(event)}

                step1 = self.define_view(u'/firstStepOfDetour', title=u'Step 1', slot_definitions=slot_with_button, detour=True)
                step2 = self.define_view(u'/lastStepOfDetour', title=u'Step 2', slot_definitions=slot_with_button)

                self.define_transition(event, step1, step2)
                self.define_return_transition(event, step2)

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                detour_ui = self.define_user_interface(u'/uiWithDetour',  UIWithDetour,  IdentityDictionary(), name=u'second_ui')
                bookmark = detour_ui.get_bookmark(relative_path='/firstStepOfDetour')
                self.define_user_interface(u'/uiWithLink',  UIWithLink,  IdentityDictionary(), name=u'first_ui', bookmark=bookmark)

                
        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open(u'/uiWithLink/initial')
        browser.click(u'//a')
        vassert( browser.location_path == u'/uiWithDetour/firstStepOfDetour' )
                
        browser.click(u'//input[@type="submit"]')
        vassert( browser.location_path == u'/uiWithDetour/lastStepOfDetour' )

        browser.click(u'//input[@type="submit"]')
        vassert( browser.location_path == u'/uiWithLink/initial' )

        # The query string is cleared after such a return (it is used to remember where to return to)
        vassert( browser.location_query_string == u'' )



# Copyright 2010-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test, set_up
from reahl.tofu import vassert, temp_dir
from reahl.stubble import stubclass, replaced

from reahl.web.dhtml import DhtmlUI
from reahl.web.fw import WebExecutionContext, UserInterface
from reahl.web.ui import Slot, TwoColumnPage
from reahl.web_dev.fixtures import WebBasicsMixin
from reahl.webdev.tools import Browser

class DjhtmlFixture(Fixture, WebBasicsMixin):
    def new_static_dir(self):
        return temp_dir()
    
    div_internals = u'<p>some stúff</p>'

    def new_dhtml_file(self):
        contents = u'<!DOCTYPE html><meta charset="UTF-8"><html><head><title>â title</title></head>' \
                   u'<body><div id="astatic">%s</div></body></html>' % self.div_internals
        return self.static_dir.file_with(u'correctfile.d.html', contents.encode('utf-8'))

    def new_afrikaans_dhtml_file(self):
        contents = u'<!DOCTYPE html><meta charset="UTF-8"><html><head><title>Afrikaans bo!</title></head><body></body></html>'
        return self.static_dir.file_with(u'correctfile.af.d.html', contents.encode('utf-8'))

    def new_other_file(self):
        return self.static_dir.file_with(u'otherfile.txt', u'other')

    @set_up
    def create_files(self):
        self.dhtml_file
        self.afrikaans_dhtml_file
        self.other_file


@istest
class BasicTests(object):
    @test(DjhtmlFixture)
    def basic_workings(self, fixture):
        """A DhtmlUI provides a UserInterface which maps to the filesystem where there may be
           a combination of .d.html and other files. When a d.html file is requested from
           it, the contents of the specified div from inside the d.html file is inserted 
           into the specified Slot. When a normal file is requested, the file is sent verbatim."""
        
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/dhtml_ui', DhtmlUI, {u'main_slot': u'main'},
                                name=u'test_ui', static_div_name=u'astatic')

        # Djhtml files should be located in the web.static_root
        fixture.config.web.static_root = fixture.static_dir.name

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI, enable_js=True)
        browser = Browser(wsgi_app)

        # A dhtml file: TwoColumnPage's main_slot now contains the insides of the div in the dhtml file
        browser.open(u'/dhtml_ui/correctfile.d.html')
        html = browser.get_html_for(u'//div[@id="bd"]/div/div/*')
        vassert( html == fixture.div_internals )

        # A non-dhtml file is returned verbatim
        browser.open(u'/dhtml_ui/otherfile.txt')
        contents = browser.raw_html
        vassert( contents == u'other' )

        # Files that do not exist are reported as such
        browser.open(u'/dhtml_ui/idonotexist.txt', status=404)
        browser.open(u'/dhtml_ui/idonotexist.d.html', status=404)

    @test(DjhtmlFixture)
    def i18n_dhtml(self, fixture):
        """Djhtml files can have i18nsed versions, which would be served up if applicable."""

        @stubclass(WebExecutionContext)
        class AfrikaansContext(WebExecutionContext):
            @property
            def interface_locale(self):
                return u'af'

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/dhtml_ui', DhtmlUI, {u'main_slot': u'main'},
                                   name=u'test_ui', static_div_name=u'astatic')

        # Djhtml files should be located in the web.static_root
        fixture.config.web.static_root = fixture.static_dir.name

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
            
        browser = Browser(wsgi_app)

        # request the file, but get the transalated alternative for the locale
        def stubbed_create_context_for_request():
            return AfrikaansContext()
        with replaced(wsgi_app.create_context_for_request, stubbed_create_context_for_request):
            browser.open(u'/dhtml_ui/correctfile.d.html')
            
        vassert( browser.title == u'Afrikaans bo!' )


